/* LILE - Lynxified Input Line Editor */
/* Copyright message is near the bottom of the file */

#include "config.h"

/* #define DEBUG /**/

#define MAX_PROMPT_LEN 60

/* special characters used by ile */
#define del '\177'
#define CA '\1'
#define CB '\2'
#define CC '\3'
#define CD '\4'
#define CE '\5'
#define CF '\6'
#define bel '\7'
#define bs '\10'
#define tab '\11' /* CI... */
#define nl '\12'
#define CK '\13'
#define CL '\14'
#define cr '\15'
#define CN '\16'
#define CO '\17'
#define CP '\20'
#define CQ '\21'
#define CR '\22'
#define CS '\23'
#define CT '\24'
#define CU '\25'
#define CV '\26'
#define CW '\27'
#define CX '\30'
#define CY '\31'
#define CZ '\32'
#define esc '\33'
#define telnesc '\35' /* telnet escape character */
 
void setup_action_table();

#include <stdio.h>
#include <fcntl.h>
#include <sgtty.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <pwd.h>
#include <utmp.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/resource.h>
/*
#include <sys/ttold.h>
*/

/* Definitions of system stuff.  */
extern int errno;

long lseek();
char *malloc();
char *realloc();
time_t time();

#ifndef TRUE
# define FALSE 0
# define TRUE  1
#endif

#define READ  0
#define WRITE 1
#define ERROR 2

#define BUFFER_SIZE 1000 /* was 255 */
#define HISTORY_SIZE 101
#define USER_NAME_SIZE 8
#define EOL (-2)

/* areas and variables used to get termcap information */
char *getenv();
char *tgetnum();
char *tgetflag();
char *tgetstr();

char termcap_entry[1024];	/* termcap entry for the users terminal */
char term_seqs[1024];		/* area to store control sequences in */
char *where = term_seqs;

char *cle;			/* move cursor left one space */
char *cce;			/* clear to end of line */
char *cbl;			/* audible bell */
char *cnl;			/* new line character */
char *ccr;			/* carriage return */
char *ccl;			/* clear screen -lynx */
char *cic;			/* insert character -lynx */
char *csc;			/* save cursor position -lynx */
char *crc;			/* restore cursor position -lynx */
char *cmd;			/* standout mode -lynx */
char *cme;			/* end standout mode -lynx */

/*
  The original tty status flags are stored so that they can be
  restored when ile exits.
*/
struct sgttyb tty_sgttyb;
struct tchars tty_tchars;
struct ltchars tty_ltchars;
struct winsize tty_winsize;
int expect_exception, ignorestop, new_prompt, output_complete;
int childpid;
int tty_ldisc;
int tty_mode;

/* file descriptors for tty and pty */
int master_pty;
int slave_tty;

/* the names of the tty and pty opened by getpty */
char ttydev[] = "/dev/ttyHI";
char ptydev[] = "/dev/ptyHO";
char prompt[MAX_PROMPT_LEN+1];

char *homedir = 0, *appl = 0, *path = 0;

/* Set up the tty in raw mode */
void init_tty() {
    struct sgttyb params;
    struct tchars tparams;
    struct ltchars ltparams;
    int mode;

    /* Simulate RAW but allow original parity to work.  Thus we use
     * CBREAK with all the options turned off. */

    params = tty_sgttyb;
    params.sg_flags = CBREAK;
    (void) ioctl(READ, TIOCSETP, &params);

    tparams = tty_tchars;
    tparams.t_intrc = -1;
    tparams.t_quitc = -1;
    tparams.t_eofc = -1;
    tparams.t_brkc = -1;
    (void) ioctl(READ, TIOCSETC, &tparams);

    ltparams = tty_ltchars;
    ltparams.t_rprntc = -1;
    ltparams.t_flushc = -1;
    ltparams.t_lnextc = -1;
    (void) ioctl(READ, TIOCSLTC, &ltparams);

    /* set new mode on tty */

    mode = LNOFLSH | LDECCTQ | LLITOUT;
    (void) ioctl(READ, TIOCLSET, &mode);
}

/* Restore tty to original mode */
void restore_tty() {
    (void) ioctl(READ, TIOCSETP, &tty_sgttyb);
    (void) ioctl(READ, TIOCSETC, &tty_tchars);
    (void) ioctl(READ, TIOCSLTC, &tty_ltchars);
    (void) ioctl(READ, TIOCLSET, &tty_mode);
}

/* clean up and leave.

This function is bound to the SIGCHLD signal so that when the
child process exits, so does ile. It is also called when an exception
is detected by select() in ile().
*/
void clean_up() {
    int pgrp;

    /* kill off the child process */
    if (slave_tty) (void) ioctl(slave_tty, TIOCGPGRP, (char *) &pgrp);
    if (pgrp) (void) killpg(pgrp, SIGTERM);
    restore_tty();

#if LOGIN
    /* "logout" the user */
    if (slave_tty) (void) setutmp(slave_tty, FALSE);
#endif

    /* clean up the tty/pty pair */
    if (master_pty) (void) close(master_pty);
    if (slave_tty) (void) close(slave_tty);

    fflush(stdout);
    exit(0);
}

/*
Forward a signal we receive to the child.  This is necessary because
the child is attached to a different terminal than we are and may not
get all the signals that we get.
Also, this implicitly synchronizes the stops/continues between parent and
child.
*/
void handle_signal(signo) {
    int pgrp;

#ifdef DEBUG
    fprintf(stderr, "forwarding signal %d to child\r\n", signo);
    fflush(stderr);
#endif
    (void) ioctl(slave_tty, TIOCGPGRP, (char *) &pgrp);
    if (pgrp && killpg(pgrp, signo) < 0) {
	perror("ile: killpg");
	clean_up();
	exit(1);
	/* NOTREACHED */
    }

    /* note the change so that we don't die after select */
    expect_exception = TRUE;
}

void handle_child();
/*
Handle a stop condition.  Restore original terminal state, then suspend
ourselves for real.
*/
void handle_stop(signo) int signo; {
    int pgrp;
    int oldmask;

#ifdef DEBUG
    fprintf(stderr, "ile stopped by signal %d\r\n", signo);
#endif

    /* make sure we are not notified of the forwarded stop */
    ignorestop = TRUE;

    /* let children stop as well */
    (void) ioctl(slave_tty, TIOCGPGRP, (char *) &pgrp);
    if (pgrp)
	(void) killpg(pgrp, SIGSTOP);

    /* restore original terminal state */
    restore_tty();
 
    (void) signal(signo, SIG_DFL);
    oldmask = sigsetmask(0);

    /* suspend ourselves */
    kill(getpid(), signo);

    (void) sigsetmask(oldmask);
    (void) signal(signo, handle_stop);

    /* put terminal back in raw mode */
    init_tty();

    /* note the change so that we don't die after select */
    expect_exception = TRUE;
}

/*
Handle a child status change.  The basic idea is to do whatever the child
did, i.e., suspend, abort, exit with the proper status etc.
*/
void handle_child() {
    union wait status;

    if (wait3(&status, WUNTRACED, NULL) != childpid) {
	fprintf(stderr, "ile: notified by unknown process\r\n");
	return;
    }

    if (WIFSTOPPED(status)) {
	/* ignore stop signals that we forwarded to the child */
	if (status.w_stopsig == SIGSTOP && ignorestop) {
	    ignorestop = FALSE;
	    return;
	}

#ifdef DEBUG
	fprintf(stderr, "child stopped by signal %d\r\n", status.w_stopsig);
#endif
	/* stop ourselves */
	handle_stop(status.w_stopsig);
    }
    else if (WIFSIGNALED(status)) {
#ifdef DEBUG
	fprintf(stderr, "child killed by signal %d\r\n", status.w_termsig);
#endif
	clean_up();

	if (status.w_coredump) {
		chdir ("/");		/* prevent own core dump */
	}
	(void) signal (status.w_termsig, SIG_DFL);
	kill (getpid(), status.w_termsig);
    }
    else {
#ifdef DEBUG
	fprintf(stderr, "child exited with %d\r\n", status.w_retcode);
#endif
	clean_up();
	exit (status.w_retcode);
    }
}

/*
  getpty opens a pty, storing file descriptors in pty and tty.
  It trys pairs in order until it finds a pair that is not in use.
*/
getpty(pty, tty) int *pty; int *tty; {
    int devindex;
    int letter;

    static char ptychar1[] = "pqrstuvwxyz";
    static char ptychar2[] = "0123456789abcdef";

    letter = 0; while (letter < 11) {
	ttydev[strlen(ttydev) - 2] = ptychar1[letter];
	ptydev[strlen(ptydev) - 2] = ptychar1[letter];
	letter++;
	devindex = 0; while (devindex < 16) {
	    ttydev[strlen(ttydev) - 1] = ptychar2[devindex];
	    ptydev[strlen(ptydev) - 1] = ptychar2[devindex];
	    devindex++;
	    if ((*pty = open(ptydev, O_RDWR)) >= 0) {
		if ((*tty = open(ttydev, O_RDWR)) >= 0) return;
		else (void) close(*pty);
	    }
	}
    }
    (void) fprintf(stderr, "ile: unable to allocate pty/tty pair\n");
    exit(1);
}

/*
Termcap entries may have a sequences of digits optionally followed
by a '*' in front of the actual sequence. This routine increments
the pointer past this information.
*/
void strip(ptr) char **ptr; {
    while (('0' <= **ptr) && (**ptr <= '9')) (*ptr)++;
    if (**ptr == '*') (*ptr)++;
}

/*
Set up everything needed to use the control sequences from the
termcap entry for the terminal.
*/
void get_termcap() {
    char *terminal_type;	/* type of terminal */

    if (!(terminal_type = getenv("TERM"))) {
	(void) fprintf(stderr, "ile: TERM not set\n");
	exit(1);
    }
    if (tgetent(termcap_entry, terminal_type) < 1) {
	(void) fprintf(stderr, "ile: bogus terminal type %s\n", terminal_type);
	exit(1);
    }
    if (cle = tgetstr("le", &where)) strip(&cle);
    else /* if (tgetflag("bs")) */ cle = "\b"; 
    if (cbl = tgetstr("bl", &where)) strip(&cbl);
    else cbl = "\7";
    if (cnl = tgetstr("nl", &where)) strip(&cnl);
    else cnl = "\n";
    if (ccl = tgetstr("cl", &where)) strip(&ccl);
    else ccl = cnl;
    if (ccr = tgetstr("cr", &where)) strip(&ccr);
    else ccr = "\r";
    if (cic = tgetstr("ic", &where)) strip(&cic);
    /* if (cce = tgetstr("cd", &where)) strip(&cce);
    else */
    if (cce = tgetstr("ce", &where)) strip(&cce);
    else {
#if 0
	(void) fprintf(stderr,
	    "ile: can't run on %s (need \"ce\" capability)\n",
	    terminal_type);
	exit(1);
#else
	cce = " @"; /* garbage is better than no ile at all */
#endif
    }
    cme = NULL;
    if (cmd = tgetstr("md", &where)) {
	if (cme = tgetstr("me", &where)) strip(&cmd), strip(&cme);
	else cmd = NULL;
    }
    crc = NULL;
    if (csc = tgetstr("sc", &where)) {
	if (crc = tgetstr("rc", &where)) strip(&csc), strip(&crc);
	else csc = NULL;
    }
#if 1
    /* special hack to obtain 'sc' and 'rc' also for xterms */
    if (!csc && !strncmp(terminal_type, "xterm", 5)) {
	if (tgetent(termcap_entry, "vt100") > 0) {
	    if (csc = tgetstr("sc", &where)) {
		if (crc = tgetstr("rc", &where)) strip(&csc), strip(&crc);
		else csc = NULL;
	    }
	}
    }
#endif
}

/* If the window changes size, tell the slave_tty about it. */
void change_window() {
    int pgrp;

    (void) ioctl(READ, TIOCGWINSZ, &tty_winsize);
    (void) ioctl(slave_tty, TIOCSWINSZ, &tty_winsize);
    (void) ioctl(slave_tty, TIOCGPGRP, (char *) &pgrp);
    if (pgrp)
	 (void) killpg(pgrp, SIGWINCH);
    /* note the change so that we don't die after select */
    expect_exception = TRUE;
}

#if LOGIN
/* set/clear the utmp slot for the pty */
int setutmp(fd, set) {
    int old0;
    int old1;
    int old2;
    int slot;
    int f;
    struct utmp ut;

    /* Must make fd's 0,1,2 correspond to slave_tty for ttyslot() to
     * function.  Ugh!  Why doesn't ttyslot() accept a fd argument?
     * 
     * save fd's */
    old0 = dup(0); old1 = dup(1); old2 = dup(2);
    if (old0 == -1 || old1 == -1 || old2 == -1) {
	perror("ile: dup");
	return (-1);		/* file table full ? */
    }
    /* set fd's 0,1,2 for ttyslot() */
    (void) dup2(fd, 0);
    (void) dup2(fd, 1);
    (void) dup2(fd, 2);
    slot = ttyslot();

    /* put the fd's back */
    (void) dup2(old0, 0);
    (void) dup2(old1, 1);
    (void) dup2(old2, 2);

    (void) close(old0);
    (void) close(old1);
    (void) close(old2);

    if (slot < 0) {
	(void) fprintf(stderr, "ile: don't know where you are\n");
	return (-1);
    }
    f = open("/etc/utmp", O_WRONLY);
    if (f == -1) return (-1);
    bzero((char *) &ut, sizeof(ut));
    if (set) {
	struct passwd *pw;
	char *cp;

	pw = getpwuid(getuid());
	if (pw == 0) {
	    (void) fprintf(stderr, "ile: who are you?\n");
	    (void) close(f);
	    return (-1);
	}
	/* skip "/dev/" */
	cp = rindex(ttydev, '/');
	if (cp == 0) cp = ttydev;
	else cp++;

	(void) strncpy(ut.ut_line, cp, sizeof(ut.ut_line));
	(void) strncpy(ut.ut_name, pw->pw_name, sizeof(ut.ut_line));
	(void) time(&ut.ut_time);
    }
    (void) lseek(f, (long) (slot * sizeof(struct utmp)), L_SET);
    (void) write(f, (char *) &ut, sizeof(ut));
    (void) close(f);
    return (0);
}
#endif

/*
Write a line to the slave_tty.
Get the slave_tty parameters, turn off echo, send the line to the
slave_tty, restore the slave_tty paramters to the way they were
before. If echo was already off, this will have no effect.
*/
write_line(line, length) char *line; int length; {
    struct sgttyb params;
    struct sgttyb new_params;

    /* get the current parameters */
    (void) ioctl(slave_tty, TIOCGETP, &params);
    new_params = params;
    new_params.sg_flags &= ~ECHO;

    /* turn off echo so we don't see the characters twice */
    (void) ioctl(slave_tty, TIOCSETP, &new_params);
    (void) write(master_pty, line, length);

    /* set the parameters back the way they were */
    (void) ioctl(slave_tty, TIOCSETN, &params);
}

/*
The editing routines are called through the edit variable. This allows
the quote and escape commands to be implemented as a straight forward
state machine instead of requiring state flags and complex switch
statements.
*/

/* line edit buffer */
static char line[BUFFER_SIZE];

static int point;		/* insertion point */
static int length;		/* total chars in buffer */

/* procedure to edit next character */
void (*edit) ();

/* history buffer */
struct {
    int length;
    char *line;
} hist[HISTORY_SIZE];

int head;			/* insertion point */
int here;			/* current displayed line */

/*
The delimiter vector is used by the forward, backward, and delete
word operations to decide that a character is a delimiter.
*/
#define CHAR_SET_SIZE 127
#define CHAR_MASK 0177
char delimit[CHAR_SET_SIZE];

/*
The action_table is used to bind sequences of keys to operations or strings.
*/
typedef enum {
    is_action, is_string
} action_type;

struct {
    action_type flag;
    union {
	void (*action) ();
	char *string;
    } aors;
} action_table[4][CHAR_SET_SIZE];

void echo();
void cic_echo();
void echoline();
void cleartoend();
void clearline();
void backspace();
void refresh();
void reprompt();
void quote_edit();
void edit_0();
void edit_1();
void edit_2();
void edit_3();
void bell();
void insert();

/*
The following routines are action routines that are executed by the
editor to carry out commands. Each routine has a single character
argument. Each routine is invoked with the character that caused it
to be invoked as its argument.

The argument isn't always useful, but it is included to provide a
consistent interface for the routines.
*/

/* Ring the bell on the terminal.  */
void bell(ch) char ch; { fputs(cbl, stdout); }

/* Pass characters to the slave. Don't mess with them at all.  */
void pass(ch) char ch; { (void) write(master_pty, &ch, 1); }

/*
Insert a character at point in the line buffer. While we are at it
update the display to show the insertion.
*/
void insert(ch) char ch; {
    int i;

    if (length >= (BUFFER_SIZE - 2)) {
    	bell('\0');
	return;
    }
    if (length == point) {
	echo(ch);
    } else {
#if 1
	if (cic) {
	    /* insert the character */
	    cic_echo(ch);
	    /* move the characters in the line buffer */
	    for (i = length; i > point; i--) line[i] = line[i - 1];
	} else
#endif
	{
	    /* display the character */
	    echo(ch);
	    if (csc) fputs(csc, stdout); /* save cursor position */
	    /* redisplay the rest of the line */
	    echoline(&line[point], (length - point));
	    /* move the characters in the line buffer */
	    /* and put the cursor back at point */
	    for (i = length; i > point; i--) {
		line[i] = line[i - 1];
		if (!crc) backspace(line[i]);
	    }
	    if (crc) fputs(crc, stdout); /* restore cursor position */
	}
    }
    /* add the character to the line buffer */
    line[point] = ch;
    /* and increment point and length */
    length++;
    point++;
}

/*
Transpose the letter under the cursor and the letter immediately to
the left of the cursor.
*/
void transpose_chars(ch) char ch; {
    char tch;
    if ((0 < point) && (point < length)) {
	/* first, update the display */
	backspace(line[point]);
	echo(line[point]);
	echo(line[point - 1]);
	/* now swap the chars in the line buffer */
	tch = line[point];
	line[point] = line[point - 1];
	line[point - 1] = tch;
	/* point moved forward one char */
	point++;
    }
}

/*
Delete a character at point in the line buffer. While we are at it
update the display to reflect the deletion.
*/
void delete_char_under(ch) char ch; {
    int i;

    if (point < length) {
	if (csc) fputs(csc, stdout); /* save cursor position */
	/* cleartoend(); */
	/* redraw the rest of the line */
	echoline(&line[point + 1], (length - point - 1));
	echo(' '); backspace(' ');
	/* build the new line */
	for (i = point + 1; i < length; i++) {
	    line[i - 1] = line[i];
	    if (!crc) backspace(line[i]);
	}
	length--;
	if (point > length) point = length;
	if (crc) fputs(crc, stdout); /* restore cursor position */
    }
}

/*
Delete the character to the left of point in the line buffer. While we
are at it update the display to reflect the deletion.
*/
void delete_char(ch) char ch; {
    int i;
    if (point > 0) {
	/* move the cursor left one character */
	backspace(line[point - 1]);
	if (csc) fputs(csc, stdout); /* save cursor position */
	/* cleartoend(); */
	/* redraw the rest of the line */
	echoline(&line[point], (length - point));
	echo(' '); backspace(' ');
	/* build the new line */
	for (i = point; i < length; i++) {
	    line[i - 1] = line[i];
	    if (!crc) backspace(line[i]);
	}
	length--;
	point--;
	if (crc) fputs(crc, stdout); /* restore cursor position */
    }
}

/*
Bind the edit vector to quote_edit so that the next character
will be placed in the line buffer.
*/
void quote(ch) char ch; { edit = quote_edit; }

/* The next character will select an action from action_table[1] */
void meta1(ch) char ch; { edit = edit_1; }

/* The next character will select an action from action_table[2] */
void meta2(ch) char ch; { edit = edit_2; }

/* The next character will select an action from action_table[3] */
void meta3(ch) char ch; { edit = edit_3; }

/* Delete the word to the left of the cursor.  */
void delete_word(ch) char ch; {
    int i;
    int old;

    if (length > 0) {
	/* find the new deletion point */
	old = point;
	/* first skip over any delimiters */
	for (; (point > 0) && (delimit[line[point - 1]]); point--)
	    backspace(line[point - 1]);
	/* now delete until we find a delimiter */
	for (; (point > 0) && (!delimit[line[point - 1]]); point--)
	    backspace(line[point - 1]);
	if (csc) fputs(csc, stdout); /* save cursor position */
	/* clear to the end of the line */
	cleartoend();
	/* redraw the rest of the line */
	echoline(&line[old], (length - old));
	/* construct the new line */
	for (i = 0; i < (length - old); i++) {
	    line[point + i] = line[old + i];
	    backspace(line[point + i]);
	}
	/* update the length */
	length = length - (old - point);
	if (crc) fputs(crc, stdout); /* restore cursor position */
    }
}

/*
Go forward one word.
*/
void forward_word(ch) char ch; {
    if (length > 0) {
	/* first skip any delimiters */
	for (; (point < length) && (delimit[line[point]]); point++)
	    echo(line[point]);
	/* now skip until we find a delimiter */
	for (; (point < length) && (!delimit[line[point]]); point++)
	    echo(line[point]);
    }
}

/* Lower case the word.  */
void lower_word(ch) char ch; {
    if (length > 0) {
	/* first skip any delimiters */
	for (; (point < length) && (delimit[line[point]]); point++)
	    echo(line[point]);
	/* now skip until we find a delimiter */
	for (; (point < length) && (!delimit[line[point]]); point++) {
	    if ((line[point] >= 'A') && (line[point] <= 'Z')) {
		line[point] = line[point] - 'A' + 'a';
		echo(line[point]);
	    } else
		echo(line[point]);
	}
    }
}

/* Upper case the word.  */
void upper_word(ch) char ch; {
    if (length > 0) {
	/* first skip any delimiters */
	for (; (point < length) && (delimit[line[point]]); point++)
	    echo(line[point]);
	/* now skip until we find a delimiter */
	for (; (point < length) && (!delimit[line[point]]); point++) {
	    if ((line[point] >= 'a') && (line[point] <= 'z')) {
		line[point] = line[point] - 'a' + 'A';
		echo(line[point]);
	    } else
		echo(line[point]);
	}
    }

}

/* Capitalize the word. */
void capitalize_word(ch) char ch; {
    if (length > 0) {
	/* first skip any delimiters */
	for (; (point < length) && (delimit[line[point]]); point++)
	    echo(line[point]);
	/* now skip until we find a delimiter */
	if ((point < length) && (!delimit[line[point]])) {
	    if ((line[point] >= 'a') && (line[point] <= 'z')) {
		line[point] = line[point] - 'a' + 'A';
		echo(line[point]);
	    } else
		echo(line[point]);
	}
	point++;
	for (; (point < length) && (!delimit[line[point]]); point++) {
	    if ((line[point] >= 'A') && (line[point] <= 'Z')) {
		line[point] = line[point] - 'A' + 'a';
		echo(line[point]);
	    } else
		echo(line[point]);
	}
    }
}

/* Go backward one word. */
void backward_word(ch) char ch; {
    if (length > 0) {
	/* first backspace over any delimiters */
	for (; (point > 0) && (delimit[line[point - 1]]); point--)
	    backspace(line[point - 1]);
	/* now backspace until we find a delimiter */
	for (; (point > 0) && (!delimit[line[point - 1]]); point--)
	    backspace(line[point - 1]);
    }
}

/* Move the cursor to the start of the line. */
void start_of_line(ch) char ch; {
    int i;

    if (length > 0) {
	for (i = 0; i < point; i++)
	    backspace(line[i]);
	point = 0;
    }
}

/* Move the cursor one character to the left.  */
void backward_char(ch) char ch; {
    if ((length > 0) && (point > 0)) {
	backspace(line[point - 1]);
	point--;
    }
}

/* Move the cursor to the right of the last character on the line. */
void end_of_line(ch) char ch; {
    if ((length > 0) && (point < length)) {
	echoline(&line[point], (length - point));
	point = length;
    }
}

/* Move the cursor one character to the right. */
void forward_char(ch) char ch; {
    if ((length > 0) && (point < length)) {
	echo(line[point]);
	point++;
    }
}

/* Add a line to the history buffer and pass it to the child process as input.
*/
void send_line(ch) char ch; {
    int prev;

    /* Put the line in the history buffer. Make here point to the current
     * line. And increment head to point to the next history slot. */

    /* If the current line is identical to the current history line, don't
     * add it. */

    /* Don't save blank lines */

    if ((head - 1) < 0) prev = HISTORY_SIZE - 1;
    else prev = head - 1;

    if ((length != 0) &&
	((length != hist[prev].length) ||
	    (strncmp(hist[prev].line, line, length) != 0))) {
	/* set the length of the entry */
	hist[head].length = length;
	/* make sure there is enough storage for the new line */
	if (hist[head].line == NULL) {
	    if ((hist[head].line = (char *) malloc((unsigned) length)) == NULL)
		perror("ile");
	} else
	    if ((hist[head].line =
		    (char *) realloc(hist[head].line, (unsigned) length))
		== NULL)
		perror("ile");

	(void) strncpy(hist[head].line, line, length);
	head = (head + 1) % HISTORY_SIZE;
	if (hist[head].line != NULL) {
	    free(hist[head].line);
	    hist[head].length = 0;
	    hist[head].line = NULL;
	}
    }
    /* reset here */
    here = head;

    /* Echo a carriage return or a newline as a cr-nl sequence. Then send the
     * line to the child process. Finally, clear the buffer for reuse. */
    fputs(ccr, stdout);
    fputs(cnl, stdout);
    line[length] = nl;
    length++;
    write_line(line, length);
    point = 0;
    length = 0;
    prompt[0] = 0;	/* erase the prompt string, hope to catch next one */
}

/* Erase the entire line.  */
void erase_line(ch) char ch; {
    /* remove any text from the display */
    clearline(ch);
    /* nothing in the line buffer */
    point = 0;
    length = 0;
    /* reset here */
    here = head;
}

/* Erase from the current cursor position to the end of the line. */
void erase_to_end_of_line(ch) char ch; {
    if ((length > 0) && (point < length)) {
	cleartoend();
	length = point;
    }
}

/* Same, but get to the next line first */
void redraw_line(ch) char ch; {
    fputs(ccr, stdout);
    fputs(cnl, stdout);
    reprompt();
}

/* Clear screen and rewrite input buffer */
void clear_screen(ch) char ch; { fputs(ccl, stdout); reprompt(); }

/*
Go to the the next entry in the history buffer and display it.
If we are past the last history entry, then beep.
*/
void forward_history(ch) char ch; {
    if (here != head) {
	clearline(ch);
	here = (here + 1) % HISTORY_SIZE;
	length = hist[here].length;
	point = length;
	(void) strncpy(line, hist[here].line, length);
	echoline(line, length);
    } else bell('\0');
}

/*
Search backward in the history list for a line that starts with
the characters left of the cursor. If it is found make it the
current line.
*/
void search_backward_history(ch) char ch; {
    int prev;
    int i;

    /* search backward in the history */
    prev = here;
    do {
	prev--;
	if (prev < 0) prev = HISTORY_SIZE - 1;
    }
    while ((hist[prev].line != NULL) &&
	(strncmp(line, hist[prev].line, point) != 0));
    /* if something was found, make it the current line */
    if (hist[prev].line != NULL) {
	/* remember the position in the history */
	here = prev;
	/* set the length and point correctly */
	length = hist[here].length;
	if (point > length) point = length;
	if (csc) fputs(csc, stdout); /* save cursor position */
	/* redraw the line */
	clearline(ch);
	(void) strncpy(line, hist[here].line, hist[here].length);
	echoline(line, length);
	if (crc) fputs(crc, stdout); /* restore cursor position */
	else for (i = point; i < length; i++) backspace(line[i]);
    }
    else bell('\0');
}

/*
Go back one entry in the history buffer and display it. If we are
already at the last entry, then beep.
*/
void backward_history(ch) char ch; {
    int prev;

    prev = here - 1;
    if (prev < 0) prev = HISTORY_SIZE - 1;
    if (hist[prev].line != NULL) {
	clearline(ch);
	here = prev;
	length = hist[here].length;
	point = length;
	(void) strncpy(line, hist[here].line, length);
	echoline(line, length);
    }
    else bell('\0');
}

/* The following routines are utility routines used by the editing routines. */

/* Clear to the end of the current input line. */
void cleartoend() {
    /* send the clear character */
    fputs(cce, stdout);
    /* send somes nulls for padding */
    fputs("\0\0\0\0", stdout);
}

/*
Clear the input line. Backspace to the start of the line. Then clear
to the end of the line.
*/
void clearline(ch) char ch; {
    int i;

    for (i = 0; i < point; i++) backspace(line[i]);
    cleartoend();
}

void cic_echo(ch) char ch; {
    if (ch == tab) printf("%s %s %s %s ", cic, cic, cic, cic);
    else if (ch < ' ') printf("%s^%s%c", cic, cic, '@'+ch);
    else printf("%s%c", cic, ch);
}

/*
Echo a character. Not all characters are created equal. Control characters
are echoed in ^X form. So they take up two character positions instead of
the normal 1 character position.
*/
void echo(ch) char ch; {
    if (ch == tab) fputs("    ", stdout); /* patch for tab char: lynx */
    else if (ch < ' ') {
	(void) fputc('^', stdout);
	(void) fputc('@' + ch, stdout);
    } else {
	(void) fputc(ch, stdout);
    }
}

/* Echo a line with echo().  */
void echoline(line, length) char *line; int length; {
    int i;

    for (i = 0; i < length; i++) echo(*line++);
}

/* Redraw the input line _with_ the prompt ahead */
void reprompt() {
    if (prompt[0]) fputs(prompt, stdout);
    if (length) refresh();
}

/* Rewrite the input line */
void refresh() {
    int i;

    if (csc && point < length) {
	/* this sort of code would be needed all over ile
	   to fix all the backspacing, but far too few
	   terminals support this.. or the termcaps don't
	   know about it...  xterm understands this, but only
	   if you switch to vt100 virtually!  :(  */
    	for (i = 0; i < length; i++) {
		if (i == point) fputs(csc, stdout); /* save cursor position */
		echo(line[i]);
	}
	fputs(crc, stdout);	/* restore cursor position */
    } else {
    	echoline(line, length);
	for (i = point; i < length; i++) backspace(line[i]);
    }
}

/*
Backspace over a character. Generate enough bs characters to backspace
over any character.
*/
void backspace(ch) char ch; {
    if (ch == tab) fputs(cle, stdout), fputs(cle, stdout); /* lynx */
    if (ch < ' ') fputs(cle, stdout);
    fputs(cle, stdout);
}

/* Add any character to the line buffer. */
void quote_edit(ch) char ch; {
    insert(ch);
    edit = edit_0;
}

/*
Given a character and an action table number either execute the
action or pass the string to (*edit)(ch)
*/
void dispatch(table, ch) int table; char ch; {
    char *cptr;

    switch (action_table[table][ch].flag) {
case is_action:
	(*(action_table[table][ch].aors.action)) (ch);
	break;
case is_string:
	cptr = action_table[table][ch].aors.string;
	while ((*cptr) != '\0') {
	    (*edit) (*cptr);
	    cptr++;
	}
	break;
    }
}

/* Select an action from action_table[3] and execute it. */
void edit_3(ch) char ch; {
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */
    edit = edit_0;
    dispatch(3, ch);
    (void) fflush(stdout);
}

/* Select an action from action_table[2] and execute it.  */
void edit_2(ch) char ch; {
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */
    edit = edit_0;
    dispatch(2, ch);
    (void) fflush(stdout);
}

/* Select an action from action_table[1] and execute it. */
void edit_1(ch) char ch; {
    /* reset so that next input is handled by edit_0 unless over ridden by
     * the action. */
    edit = edit_0;
    dispatch(1, ch);
    (void) fflush(stdout);
}

/* Select an action from action_table[0] and execute it.  */
void edit_0(ch) char ch; {
    if (ch == tty_tchars.t_eofc && point == 0 && length == 0) {
	int mode = 1;
#ifdef DEBUG
	fprintf(stderr, "EOF received\r\n");
#endif
	(void) ioctl(master_pty, TIOCREMOTE, &mode);
	write(master_pty, NULL, 0);
	mode = 0;
	(void) ioctl(master_pty, TIOCREMOTE, &mode);
    }
    else {
	dispatch(0, ch);
    }

    (void) fflush(stdout);
}

/* Input line editor.
Initialize the world. Then loop forever using select to wait for
characters to be available from either stdin or from master_pty.
When characters are available, pass them on after doing any needed
editing.
*/
void ile() {
    /* general purpose integer variable */
    int i;
    /* arguments for read and write calls */
    char buffer[BUFFER_SIZE];
    int cc;
    /* current slave_tty parameters */
    struct sgttyb slave_params;
    /* Arguments for select call */
    int nfds;
    int width;
    int readfds;

    /* Get all the different pieces of the current ttys' state and copy them
     * to the slave_tty. */
    /* tty sgttyb */
    (void) ioctl(READ, TIOCGETP, &tty_sgttyb);
    (void) ioctl(slave_tty, TIOCSETP, &tty_sgttyb);
    /* tty line discipline */
    (void) ioctl(READ, TIOCGETD, &tty_ldisc);
    (void) ioctl(slave_tty, TIOCSETD, &tty_ldisc);
    /* tty tchars */
    (void) ioctl(READ, TIOCGETC, &tty_tchars);
    (void) ioctl(slave_tty, TIOCSETC, &tty_tchars);
    /* tty mode */
    (void) ioctl(READ, TIOCLGET, &tty_mode);
    (void) ioctl(slave_tty, TIOCLSET, &tty_mode);
    /* tty ltchars */
    (void) ioctl(READ, TIOCGLTC, &tty_ltchars);
    (void) ioctl(slave_tty, TIOCSLTC, &tty_ltchars);
    /* tty windsize */
    (void) ioctl(READ, TIOCGWINSZ, &tty_winsize);
    (void) ioctl(slave_tty, TIOCSWINSZ, &tty_winsize);
    expect_exception = FALSE;
#if LOGIN
    /* "login" the user */
    (void) setutmp(slave_tty, TRUE);
#endif
    /* set up new tty state */
    init_tty();

    /* what to do if the child or parent dies */
    (void) signal(SIGCHLD, handle_child);

    (void) signal(SIGTSTP, handle_stop);
    (void) signal(SIGTTIN, handle_stop);
    (void) signal(SIGTTOU, handle_stop);

    (void) signal(SIGHUP, handle_signal);
    (void) signal(SIGINT, handle_signal);
    (void) signal(SIGQUIT, handle_signal);
    (void) signal(SIGTERM, handle_signal);
    (void) signal(SIGCONT, handle_signal);
    (void) signal(SIGUSR1, handle_signal);
    (void) signal(SIGUSR2, handle_signal);

    /* what to do it the window changes size */
    (void) signal(SIGWINCH, change_window);

    /* get descriptor table size */
    width = getdtablesize();
    if (width > 32) width = 32;
    /* set initial edit function */
    edit = edit_0;
    /* initialize line buffer */
    point = 0; length = 0;
    /* initialize history buffer */
    head = 0; here = 0;
    for (i = 0; i < HISTORY_SIZE; i++) {
	hist[i].length = 0;
	hist[i].line = NULL;
    }

    for (;;) {
	fflush(stdout);
	readfds = (1 << READ) | (1 << master_pty);
	/* wait for input from stdin or master_pty */
	nfds = select(width,
	    (fd_set *) & readfds,
	    (fd_set *) NULL,
	    (fd_set *) NULL,
	    (struct timeval *) NULL);
	if (nfds == -1) {		/* an exception has occured */
	    if (expect_exception) {
		/* nothing serious, the window changed size */
		expect_exception = FALSE;
	    } else {
		perror("ile: received exception");
		clean_up();
	    }
	}
	else if ((nfds > 0) && (readfds != 0)) { /* something to read */
	    if ((readfds & (1 << master_pty)) != 0) {
#if SYNC
		if (output_complete) {
		    if (length) clearline();
		    fputs(ccr, stdout);
		    fputs(cce, stdout);
		    fflush(stdout);
		}
#endif
		/* read the pending characters. */
		cc = read(master_pty, buffer, BUFFER_SIZE);
		/* display the characters. */
		(void) write(WRITE, buffer, cc);
		/* prompt recognition logic is pretty guessy! */
		if (cc>1 && buffer[cc-1]==' ' && buffer[cc-2]!=' '
			&& !isalpha(buffer[cc-2])) {
		    int l;
		    for(i = cc-2; i>=0; i--)
		     if (buffer[i] == cnl[0]) break;
		    i++;	/* first char of presumed prompt */
		    l = cc-i;	/* length of presumed prompt */
		    if (l>1 && l<MAX_PROMPT_LEN) {
		    	strncpy(prompt, &buffer[i], l);
			prompt[l] = 0;
#if SYNC
			new_prompt = 1;
#endif
#ifdef DEBUG
			printf("\n[prompt set to '%s']\n", prompt);
#endif
		    }
		}
#if SYNC
		if (new_prompt) {
		    new_prompt = 0;
		    output_complete = 1;
		    if (length) refresh();
		} else {
		    /* again.. rather guessy */
		    if (cc && buffer[cc-1]=='\n') {
			reprompt();
		    	output_complete = 1;
		    } else
			output_complete = 0; /* in the middle of output! */
		}
#endif
	    }
	    if ((readfds & (1 << READ)) != 0) {
		/* read the pending characters. */
		cc = read(READ, buffer, BUFFER_SIZE);

		/* if the slave is in RAW or CBREAK mode, or has turned off
		 * ECHO then we should not mess with its input characters */
		(void) ioctl(slave_tty, TIOCGETP, &slave_params);
		if (((slave_params.sg_flags & (RAW | CBREAK)) != 0) ||
		    (slave_params.sg_flags & ECHO) == 0) {
		    edit = pass;
#if SYNC
		    output_complete = 0;
#endif
		}
		else if (edit == (void (*) ()) pass) {
		    edit = edit_0;
		}
		/* decide what to do with the characters. */
		for (i = 0; i < cc; i++)
		    (*edit) (CHAR_MASK & buffer[i]);
	    }
	}
    }
}

/* The child process.
Make the pty the processes controling terminal. Bind the pty to
stdin, stdout, and stderr. Then exec the users program.
*/
void child(argv, index) char *argv[]; int index; {
    /* close all file descriptors */
    (void) close(READ);
    (void) close(WRITE);
    (void) close(ERROR);
    (void) close(slave_tty);
    /* get rid of controlling terminal */
    {
	int tty;

	if ((tty = open("/dev/tty", O_RDWR) == -1) ||
	    (ioctl(0, TIOCNOTTY, 0) == -1) || (close(tty) == -1)) {
	    perror("ile: control tty");
	}
    }
    /* open the tty again, this makes the pty the controlling terminal */
    if ((slave_tty = open(ttydev, O_RDWR)) == -1) perror("ile: slave tty");
    /* slave_tty is now stdin, bind slave_tty to stdout */
    (void) dup2(slave_tty, WRITE);
    /* bind slave_tty to stderr */
    (void) dup2(slave_tty, ERROR);
    /* close master_pty descriptor */
    (void) close(master_pty);

    if (index && argv[index]) {
	/* execute specified program */
	execvp(argv[index], &argv[index]);
    } else {
	char *dashname;

	/* if the current argv[0] starts with -, then the new argv[0] */
	/* should start with - */
	if (*(argv[0]) == '-') {
	    dashname = (char *) malloc((unsigned) strlen(appl) + 2);
	    (void) strcpy(dashname, "-");
	    (void) strcat(dashname, appl);
	} else
	    dashname = appl;
	execlp(path, dashname, 0);
    }
    /* this executes if exec fails */
    perror("ile: exec");
    exit(1);
}

/* Set up default key bindings and delimeters. */
void default_bindings() {
    int i;

    /* clear delimiter vector and the action table */
    for (i = 0; i < CHAR_SET_SIZE; i++) {
	delimit[i] = FALSE;

	action_table[0][i].aors.action = insert;
	action_table[1][i].aors.action = bell;
	action_table[2][i].aors.action = bell;
	action_table[3][i].aors.action = bell;

	action_table[0][i].flag = is_action;
	action_table[1][i].flag = is_action;
	action_table[2][i].flag = is_action;
	action_table[3][i].flag = is_action;
    }
    /* default delimiters */
    delimit[' '] = TRUE;	/* blank */
    delimit['/'] = TRUE;	/* slash */
    delimit['.'] = TRUE;	/* dot */
    delimit['-'] = TRUE;	/* dash */

    setup_action_table();
}

#if ILERC
/*
Return a character or EOF. This routine reads characters from input
and converts them into a character using the following rules.

The character may be a single character, a control
character indicated by ^x, an octal number starting with \, or an
escaped character indictated by \x.
*/
int scan_char(input) FILE *input; {
    int ch;
    int value;

    ch = fgetc(input);
    switch (ch) {
case '^':
	/* it is a control character */
	for (ch = fgetc(input); '@' <= ch; ch = ch - '@');
	break;
case '\\':
	/* octal or an escaped character? */
	ch = fgetc(input);
	if (('0' <= ch) && (ch <= '7')) {
	    /* its an octal number */
	    value = 0;
	    while (('0' <= ch) && (ch <= '7')) {
		value = (value * 8) + (ch - '0');
		ch = fgetc(input);
	    }
	    (void) ungetc(ch, input);
	    ch = value & 0177;	/* make sure it is in range */
	}
	else {
	    /* its an escaped character */
	    ch = fgetc(input);
	}
	break;
case '\n':
	/* the real end of the line */
	ch = EOL;
	break;
default:
	/* it is just itself */
	break;
    }
    return (ch);
}

/* Set key bindings and delimiters from the users file. */
void user_bindings(file) FILE *file; {
# define NAME_SIZE 40
    static struct action_name_table {
	char *name;
	void (*action) ();
    } action_name_table[] = {
	{ "bell", bell },
	{ "clear_screen", clear_screen },
	{ "pass", pass },
	{ "insert", insert },
	{ "transpose_chars", transpose_chars },
	{ "delete_char", delete_char },
	{ "delete_char_under", delete_char_under },
	{ "quote", quote },
	{ "meta1", meta1 },
	{ "meta2", meta2 },
	{ "meta3", meta3 },
	{ "delete_word", delete_word },
	{ "upper_word", upper_word },
	{ "lower_word", lower_word },
	{ "capitalize_word", capitalize_word },
	{ "forward_word", forward_word },
	{ "backward_word", backward_word },
	{ "start_of_line", start_of_line },
	{ "backward_char", backward_char },
	{ "end_of_line", end_of_line },
	{ "forward_char", forward_char },
	{ "send_line", send_line },
	{ "erase_line", erase_line },
	{ "erase_to_end_of_line", erase_to_end_of_line },
	{ "redraw_line", redraw_line },
	{ "forward_history", forward_history },
	{ "backward_history", backward_history },
	{ "search_backward_history", search_backward_history },
	{ "", NULL }
    };
    char name[NAME_SIZE];
    int ch, i;
    int linecount;
    int table;
    int entry;

    /* First clear the default delimiters */
    for (i = 0; i < CHAR_SET_SIZE; i++) delimit[i] = FALSE;
    /* Now read the delimiter characters */
    while (((int) (ch = fgetc(file)) != EOF) && (ch != '\n'))
	delimit[ch] = TRUE;

    linecount = 2;
    /* Now read the character binding pairs */
    while ((int) (ch = fgetc(file)) != EOF) {
	switch (ch) {
case '\n':
	    /* skipping a blank line */
	    linecount++;
	    break;
case '0':
case '1':
case '2':
case '3':
	    /* which table is this entry directed to? */
	    table = ch - '0';
	    /* get the character code */
	    entry = scan_char(file);
	    /* make sure the '=' is there */
	    ch = fgetc(file);
	    if (ch != '=') {
		(void) fprintf(stderr, "ile: '=' missing on line %d\n",
		    linecount);
		exit(1);
	    }
	    /* collect the action name or string */
	    for (ch = scan_char(file), i = 0;
		((int) ch != EOL) && (i < (NAME_SIZE - 1));
		ch = scan_char(file), i++)
	    {
		name[i] = ch;
		name[i + 1] = '\0';
	    }
	    /* look it up in the action_name_table */
	    for (i = 0;
		(action_name_table[i].action != NULL) &&
		(strcmp(name, action_name_table[i].name) != 0);
		i++);
	    /* if it was found, put it in the action array */
	    if (action_name_table[i].action == NULL) {
		/* must be a string */
		action_table[table][entry].flag = is_string;
		action_table[table][entry].aors.string =
		    (char *) malloc((unsigned) strlen(name) + 1);
		(void) strcpy(action_table[table][entry].aors.string, name);
	    } else {
		/* its an action */
		action_table[table][entry].flag = is_action;
		action_table[table][entry].aors.action =
		    action_name_table[i].action;
	    }
	    linecount++;	/* count the line */
	    break;
default:
	    (void) fprintf(stderr,
		"\nile: error in initialization file on line %d\n",
		linecount);
	    exit(1);
	}
    }
    (void) fclose(file);
}

FILE* get_rc_file(suggestion) char *suggestion; {
    FILE *file;
    char name[BUFFER_SIZE];
    char *pwd;
    int cut;

    /* Look for an initialization file. If it's there, load it. */
    if (suggestion && (file = fopen(suggestion, "r"))) return file;
    homedir = getenv("HOME");
    if (homedir == NULL) {
	name[0] = '\0'; /* no home dir, use / instead */
    } else {
	(void) strcpy(name, homedir);
    }
    (void) strcat(name, "/.ilerc");
    cut = strlen(name);
    name[cut] = '-';
    name[cut+sizeof(char)] = '\0';
    (void) strcat(name, appl);
    if ((file = fopen(name, "r")) || (file = fopen("./.ilerc", "r")))
	return file;
    name[cut] = '\0';
    if (file = fopen(name, "r")) return file;
    return NULL;
}
#endif

main(argc, argv) int argc; char *argv[]; {
    int index;

    /* get control sequences from termcap */
    get_termcap();

#if DOCLS
    clear_screen();
#endif
#ifdef BANNER
    if (cmd) fputs(cmd, stdout);
    (void) fputs(BANNER, stdout);
    if (cme) fputs(cme, stdout);
    fputs(ccr, stdout);
    fputs(cnl, stdout);
#endif

    /* create the tty/pty pair */
    getpty(&master_pty, &slave_tty);

    /* set up the default bindings */
    default_bindings();

#if ILERC
    if (argv[1] && (*argv[1] == '-')) {
	path = argv[index = 2];
    } else {
	path = argv[index = 1];
    }
#else
    path = argv[index = 1];
#endif
    if (!path) {
	/* get the name of the users shell. default to /bin/csh */
	if ((path = getenv("SHELL")) == NULL || (*path == '\0')) {
	    path = "/bin/csh";
	}
	index = 0;
    }
    /* get just the name */
    if ((appl = strrchr(path, '/')) != NULL) appl += sizeof(char);
    else appl = path;
#if ILERC
    {
	FILE* f = get_rc_file(index == 2 ? argv[1]+sizeof(char) : NULL);
	if (f) user_bindings(f);
    }
#endif
    /* create the child process */
    childpid = fork();

    switch (childpid) {
case 0:			/* child process */
	child(argv, index);
	break;
case -1:			/* fork failed */
	perror("ile: fork");
	exit(1);
default:			/* parent process */
	ile();
	break;
    }
}

/*
                          COPYRIGHT 1988
              Evans & Sutherland Computer Corporation
                        Salt Lake City, Utah
                        All Rights Reserved.

     THE INFORMATION  IN  THIS  SOFTWARE  IS  SUBJECT  TO  CHANGE
     WITHOUT  NOTICE  AND SHOULD NOT BE CONSTRUED AS A COMMITMENT
     BY  EVANS  &  SUTHERLAND.   EVANS  &  SUTHERLAND   MAKES  NO
     REPRESENTATIONS  ABOUT  THE SUITABILITY OF THIS SOFTWARE FOR
     ANY PURPOSE.  IT IS SUPPLIED  "AS  IS"  WITHOUT  EXPRESS  OR
     IMPLIED WARRANTY.

     IF THE SOFTWARE IS MODIFIED IN A MANNER CREATING  DERIVATIVE
     COPYRIGHT  RIGHTS,  APPROPRIATE LEGENDS MAY BE PLACED ON THE
     DERIVATIVE WORK IN ADDITION TO THAT SET FORTH ABOVE.

     Permission  to  use,  copy,  modify,  and  distribute   this
     software  and  its documentation for any purpose and without
     fee is hereby granted, provided  that  the  above  copyright
     notice  appear  in  all  copies  and that both the copyright
     notice and this permission notice appear in supporting docu-
     mentation,  and  that  the name of Evans & Sutherland not be
     used in advertising or publicity pertaining to  distribution
     of the software without specific, written prior permission.

Written by:

                        Robert C. Pendleton <bobp@hal.com>

Grateful acknowledgement is made of code and ideas contributed by

 	Ian Donaldson,
 	Department of Communications & Electronic Engineering,
 	Royal Melbourne Institute of Technology,
 	Melbourne, Australia.

Old $Header: ile.c,v 2.3 88/11/11 10:31:56 bpendlet Exp $

lynxified ile, release one:
    - Added #ifdef's for ILERC and suchlike
    - Added key functions clear_screen, delete_under_or_pass
    - General mess up of source code
    - Removed the csh-specific file completion hack
    - Adapted for networking oriented use (telnet, ftp)

lile, release two:
    - Added Stolcke's magnificent signal handling code
    - Removed delete_under_or_pass, replaced by Andy's EOF recognition patch
    - Added prompt recognition (nice for CTRL-L and CTRL-R)
    - Added 'sc'/'rc' display technique since backspacing seldom works
	across line wraps
    - Added 'ic' inserting ability, again to avoid backspacing

Thanks to Andreas Stolcke <stolcke@ICSI.Berkeley.EDU> for contributing
proper signal handling functionality. Finally a CTRL-Z will do what it's
supposed to!

*/

void setup_action_table() {
    /* default action_table[0] */
    action_table[0][CA].aors.action = start_of_line;
    action_table[0][CB].aors.action = backward_char;
    action_table[0][CE].aors.action = end_of_line;
    action_table[0][CF].aors.action = forward_char;
    action_table[0][CK].aors.action = erase_to_end_of_line;
    action_table[0][CU].aors.action = erase_line;
    action_table[0][CR].aors.action = redraw_line;
    action_table[0][CN].aors.action = forward_history;
    action_table[0][CP].aors.action = backward_history;
    action_table[0][CO].aors.action = search_backward_history;
    action_table[0][CT].aors.action = transpose_chars;
    action_table[0][CV].aors.action = quote;
    action_table[0][bs].aors.action = delete_char;
    action_table[0][del].aors.action = delete_char;
    action_table[0][esc].aors.action = meta1;
    action_table[0][cr].aors.action = send_line;
    action_table[0][nl].aors.action = send_line;
    /* action_table[0][CX].aors.action = meta2; */
    action_table[0][CC].aors.action = pass;
    action_table[0][CD].aors.action = delete_char_under;
    action_table[0][CQ].aors.action = pass;
    action_table[0][CL].aors.action = clear_screen;
    action_table[0][CS].aors.action = pass;
    action_table[0][CZ].aors.action = pass;
    action_table[0][telnesc].aors.action = pass;
    /* default action_table[1] ^[ c */
    action_table[1]['b'].aors.action = backward_word;
    action_table[1]['f'].aors.action = forward_word;
    action_table[1]['k'].aors.action = erase_line;
    action_table[1][del].aors.action = delete_word;
    action_table[1][bs].aors.action = delete_word;
    action_table[1]['u'].aors.action = upper_word;
    action_table[1]['l'].aors.action = lower_word;
    action_table[1]['c'].aors.action = capitalize_word;
    action_table[1]['['].aors.action = meta2;
    action_table[1]['O'].aors.action = meta2;
    /* default action_table[2] ^[ [ */
    action_table[2]['A'].aors.action = backward_history;
    action_table[2]['B'].aors.action = forward_history;
    action_table[2]['C'].aors.action = forward_char;
    action_table[2]['D'].aors.action = backward_char;
}
